% treeList.pro: Restyled copy of theory Tree.pl by R.A.O'Keefe, 
% 25 November 1983.
% Updatable binary trees in list form.

%/*  These are the routines I meant to describe in DAI-WP-150, but the
%    wrong version went in.  We have
%	list__to_treeList    : O(N)
%	treeList__to_list    : O(N)
%	treeList__has_size   : O(N)
%	pred__maps_treeList__into_treeList : O(N)
%	newTreeList          : O(lg N)
%    where N is the number of elements in the treeList.
%    The way index__in_treeList__eq_label
%    and newTreeList work is worth noting: they build up a pattern which
%    is matched against the whole treeList when the position number finally
%    reaches 1.  In effect they start out from the desired node and
%    build up a path to the root.  They still cost O(lg N) time rather
%    than O(N) because the patterns contain O(lg N) distinct variables,
%    with no duplications.  newTreeList simultaneously builds up a pattern
%    to match the old treeList and a pattern to match the new treeList.
%*/
%
%:- public
%	index__in_treeList__eq_label/3,
%	list__to_treeList/2,
%	pred__maps_treeList__into_treeList/3,
%	newTreeList/4,
%	treeList__has_size/2,
%	treeList__to_list/2.
%
%:- mode
%	index__in_treeList__eq_label(+, +, ?),
%	index__in_treeList__eq_nodeTerm(+, +, +),
%	list__to_treeList(+, -),
%	list__to_treeList(+, +, -),
%	list__to_treeList(+),
%	pred__maps_treeList__into_treeList(+, +, -),
%	newTreeList(+, +, +, -),
%	index__in_treeList__eq_nodeTerm(+, +, +, -, +),
%	treeList__has_size(+, ?),
%	treeList__has_size(+, +, -),
%	treeList__to_list(+, -),
%	treeList__to_list(+, -, -).
%

%   index__in_treeList__eq_label(Index, treeList, Label)
%   treats the treeList as an array of N elements and returns the Index-th.
%   If Index < 1 or > N it simply fails, there is no such element.

index__in_treeList__eq_label(N, treeList, Label) :-
	index__in_treeList__eq_nodeTerm(N, treeList, t(Label,_,_)).


	index__in_treeList__eq_nodeTerm(1, treeList, treeList) :-
	   !.
	index__in_treeList__eq_nodeTerm(N, treeList, Node) :-
		N > 1,
		0 is N mod 2,
		M is N  /  2,
		!,
		index__in_treeList__eq_nodeTerm(M, treeList, t(_,Node,_)).
	index__in_treeList__eq_nodeTerm(N, treeList, Node) :-
		N > 2,
		1 is N mod 2,
		M is N  /  2,
		!,
		index__in_treeList__eq_nodeTerm(M, treeList, t(_,_,Node)).



%   list__to_treeList(List, treeList)
%   takes a given List of N elements and constructs a binary treeList
%   where index__in_treeList__eq_label(K, treeList, Lab) <=> Lab is the Kth element of List.

list__to_treeList(List, treeList) :-
	list__to_treeList(List, [treeList|Tail], Tail).


	list__to_treeList([Head|Tail], [t(Head,Left,Right)|Qhead],[Left,Right|Qtail]) :-
		list__to_treeList(Tail, Qhead, Qtail).
	list__to_treeList([], Qhead, []) :-
		list__to_treeList(Qhead).
	

		list__to_treeList([t|Qhead]) :-
			list__to_treeList(Qhead).
		list__to_treeList([]).



%   pred__maps_treeList__into_treeList(Pred, OldtreeList, NewtreeList)
%   is true when OldtreeList and NewtreeList are binary treeLists of the same shape
%   and Pred(Old,New) is true for corresponding elements of the two treeLists.
%   In fact this routine is perfectly happy constructing either treeList given
%   the other, I have given it the mode I have for that bogus reason
%   "efficiency" and because it is normally used this way round.  This is
%   really meant more as an illustration of how to map over treeLists than as
%   a tool for everyday use.

pred__maps_treeList__into_treeList(Pred, t(Old,OLeft,ORight), t(New,NLeft,NRight)) :-
	apply(Pred, [Old,New]),
	pred__maps_treeList__into_treeList(Pred, OLeft, NLeft),
	pred__maps_treeList__into_treeList(Pred, ORight, NRight).
pred__maps_treeList__into_treeList(_, t, t).



%   newTreeList(Index, OldtreeList, Label, NewtreeList)
%   constructs a new treeList the same shape as the old which moreover has the
%   same elements except that the Index-th one is Label.  Unlike the
%   "arrays" of Arrays.Pl, OldtreeList is not modified and you can hang on to
%   it as long as you please.  Note that O(lg N) new space is needed.

newTreeList(N, Old, Label, New) :-
	index__in_treeList__eq_nodeTerm(N, Old, t(_,Left,Right), New, t(Label,Left,Right)).


	index__in_treeList__eq_nodeTerm(1, Old, Old, New, New) :-
	!.
	index__in_treeList__eq_nodeTerm(N, Old, OldSub, New, NewSub) :-
		N > 1,
		0 is N mod 2,
		M is N  /  2,
		!,
		index__in_treeList__eq_nodeTerm(M, Old, t(Label,OldSub,Right),
		New,
		t(Label,NewSub,Right)).
	index__in_treeList__eq_nodeTerm(N, Old, OldSub, New, NewSub) :-
		N > 2,
		1 is N mod 2,
		M is N  /  2,
		!,
		index__in_treeList__eq_nodeTerm(M, Old, t(Label,Left,OldSub),
		New,
		t(Label,Left,NewSub)).



%   treeList__has_size(treeList, Size)
%   calculates the number of elements in the treeList.  All treeLists made by
%   list__to_treeList that are the same size have the same shape.

treeList__has_size(treeList, Size) :-
	treeList__has_size(treeList, 0, Total),
	!,
	Size = Total.


	treeList__has_size(t(_,Left,Right), SoFar, Total) :-
		treeList__has_size(Right, SoFar, M),
		N is M+1,
		!,
		treeList__has_size(Left, N, Total).
	treeList__has_size(t, Accum, Accum).



%   treeList__to_list(treeList, List)
%   is the converse operation to list__to_treeList.  Any mapping or checking
%   operation can be done by converting the treeList to a list, mapping or
%   checking the list, and converting the result, if any, back to a treeList.
%   It is also easier for a human to read a list than a treeList, as the
%   order in the treeList goes all over the place.

treeList__to_list(treeList, List) :-
	treeList__to_list([treeList|Tail], Tail, List).


	treeList__to_list([], [], []) :-
	!.
	treeList__to_list([t|_], _, []) :-
	!.
	treeList__to_list([t(Head,Left,Right)|Qhead],[Left,Right|Qtail], [Head|Tail]) :-
		treeList__to_list(Qhead, Qtail, Tail).


