% REAL.PRO: Real number operators.
% Courtesy the author of the Wumpus World Simulator.
% Edited for matching IPL naming conventions by Frans Meijs.
%
%   real__isAbsOf_real(Y,X): Y is the absolute value of X.
%   real__isSqrtOf_real(Y,X): Y is the square root of X.
%   real__isExpOf_real(Y,X): Y is e^X.
%   real__isLogOf_real(Y,X): Y is logarithm base e of X (X >= 0).
%   real__isLog2Of_real(Y,X): Y is logarithm base 2 of X (X >= 0).
%   real__isLog10Of_real(Y,X): Y is logarithm base 10 of X (X >=0).
%   real__isCosOf_real(Y,X): Y is the cosine of X (X in radians).
%   real__isSinOf_real(Y,X): Y is the sine of X (X in radians).
%   real__isArctanOf_real(Y,X,Z): Z (-pi < Z <= pi radians) is inverse tangent of the
%                  point (Y,X) (X and Y are any real number).



real__isAbsOf_real(X,X) :-
  X >= 0.
real__isAbsOf_real(Y,X) :-
  X < 0,
  Y is - X.


real__isSqrtOf_real(Y,X) :-
  sqrt1(Y,X,1.0).


sqrt1(Z,X,Z) :-
  W is ((Z * Z) - X),
  real__isAbsOf_real(AW,W),
  sqrtPrecision(P),
  AW < P,
  !.
sqrt1(Y,X,Z) :-
  Z1 is (Z - (((Z * Z) - X) / (2 * Z))),
  sqrt1(Y,X,Z1).


real__isExpOf_real(Y,X) :-
  X1 is float(X),
  expSeries(X1,1,Y1,1),
  Y is Y1 + 1.0.


% expSeries(X,Term,Y,Exp): Y is e raised to the X power according
%   to the series:
%
%   e^X = 1 + X + X^2/2! + X^3/3! + X^4/4! + ...
%
%      where (2,3,4,...) continues until exceeding expSeriesLimit.
expSeries(_,_,0,Exp) :-
  expSeriesLimit(L),
  Exp > L,
  !.
expSeries(X,Term,Y,Exp) :-
  Term1 is Term * X / Exp,
  Exp2 is Exp + 1,
  expSeries(X,Term1,Y2,Exp2),
  Y is Term1 + Y2.


real__isLog2Of_real(Y,X) :-
  real__isLogOf_real(Ye,X),
  Y is Ye / 0.69314718.  % log_e(2) = 0.69314718


real__isLog10Of_real(Y,X) :-
  real__isLogOf_real(Ye,X),
  Y is Ye / 2.302585093.  % log_e(10) = 2.302585093


real__isLogOf_real(0,0).
real__isLogOf_real(Y,X) :-
  X > 0,
  logSeriesLimit(L),
  log_eSeries(X,1,Y,1,L).


% log_eSeries(X,Term,Y,Exp,Limit): Y is the natural logarithm (base e)
%   of X (X >= 0) according to the logarithmic series:
%
%   log_e (X) = 2 [ Term + (1/3)*Term^3 + (1/5)*Term^5 + ... +
%                          (1/Exp)*Term^Exp ]
%
%      where Term = ((X-1)/(X+1)) and Exp is the largest odd <= Limit.
log_eSeries(_,_,0,Exp,Limit) :-
  Exp > Limit,
  !.
log_eSeries(X,_,Y,1,Limit) :-
  !,
  Term1 is ((X-1)/(X+1)),
  Term2 is Term1 * 2,
  log_eSeries(X,Term1,Y2,3,Limit),
  Y is Term2 + Y2.
log_eSeries(X,Term,Y,Exp,Limit) :-
  Term1 is Term * ((X-1)/(X+1)) * ((X-1)/(X+1)),
  Term2 is Term1 * (2 / Exp),
  Exp2 is Exp + 2,
  log_eSeries(X,Term1,Y2,Exp2,Limit),
  Y is Term2 + Y2.


real__isCosOf_real(Y,X) :-
  cosSeries(Y,X,1).


% cosSeries(Y,X,Base): Expands the following series to compute Y = real__isCosOf_real(X):
%
%   real__isCosOf_real(X) = (1 - 4X^2/Pi^2)*(1 - 4x^2/(3Pi)^2)*(1 - 4x^2/(5Pi)^2)*...
%
%     until the Base (1,3,5,...) exceeds cosSeriesLimit(L).
cosSeries(1.0,_,Base) :-
  cosSeriesLimit(Limit),
  Base > Limit,
  !.
cosSeries(Y,X,Base) :-
  pi(Pi),
  Term1 is (2.0 * X) / (Base * Pi),
  Term2 is 1.0 - (Term1 * Term1),
  Base1 is Base + 2,
  cosSeries(Y1,X,Base1),
  Y is Term2 * Y1.


real__isSinOf_real(Y,X) :-
  sinSeries(Y,X,1).


% sinSeries(Y,X,Base): Expands the following series to compute Y = real__isSinOf_real(X):
%
%   real__isSinOf_real(X) = (1 - X^2/Pi^2)*(1 - x^2/(2Pi)^2)*(1 - x^2/(3Pi)^2)*...
%
%     until the Base (1,2,3,...) exceeds sinSeriesLimit(L).
sinSeries(X,X,Base) :-
  sinSeriesLimit(Limit),
  Base > Limit,
  !.
sinSeries(Y,X,Base) :-
  pi(Pi),
  Term1 is X / (Base * Pi),
  Term2 is 1.0 - (Term1 * Term1),
  Base1 is Base + 1,
  sinSeries(Y1,X,Base1),
  Y is Term2 * Y1.


real__isArctanOf_real(Z,X,Y) :-
  FX is float(X),
  FY is float(Y),
  arctan1(Z,FX,FY).


% arctan1(Z,X,Y) :- Computes Z as the inverse tangent at (Y,X) based on the
%   quadrant of (Y,X) and their ratio.
arctan1(Z,0.0,Y) :-
  Y > 0.0,
  !,
  pi(Pi),
  Z is Pi / 2.0.
arctan1(Z,0.0,Y) :-
  Y < 0.0,
  !,
  pi(Pi),
  Z is - (Pi / 2.0).
arctan1(0.0,X,0.0) :-
  X >= 0.0,
  !.
arctan1(Z,X,0.0) :-
  X < 0.0,
  !,
  pi(Z).
arctan1(Z,X,Y) :-
  X > 0.0,          % 1st quadrant
  Y > 0.0,
  X >= Y,           % angle =< 45 degrees
  !,
  Ratio is Y / X,
  arctan2(Z,Ratio).
arctan1(Z,X,Y) :-
  X > 0.0,          % 1st quadrant
  Y > 0.0,
  X < Y,            % angle > 45 degrees
  Ratio is X / Y,
  !,
  arctan2(Z1,Ratio),
  pi(Pi),
  Z is (Pi / 2.0) - Z1.
arctan1(Z,X,Y) :-
  X < 0.0,          % 2nd quadrant
  Y > 0.0,
  Y =< - X,         % angle >= 135 degrees
  !,
  Ratio is - Y / X,
  arctan2(Z1,Ratio),
  pi(Pi),
  Z is Pi - Z1.
arctan1(Z,X,Y) :-
  X < 0.0,          % 2nd quadrant
  Y > 0.0,
  Y > - X,          % angle < 135 degrees
  !,
  Ratio is - X / Y,
  arctan2(Z1,Ratio),
  pi(Pi),
  Z is (Pi / 2.0) + Z1.
arctan1(Z,X,Y) :-
  X < 0.0,          % 3rd quadrant
  Y < 0.0,
  X =< Y,           % angle =< 225 degrees
  !,
  Ratio is Y / X,
  arctan2(Z1,Ratio),
  pi(Pi),
  Z is Z1 - Pi.
arctan1(Z,X,Y) :-
  X < 0.0,          % 3rd quadrant
  Y < 0.0,
  X > Y,            % angle > 225 degrees
  !,
  Ratio is X / Y,
  arctan2(Z1,Ratio),
  pi(Pi),
  Z is - (Pi / 2.0) - Z1.
arctan1(Z,X,Y) :-
  X > 0.0,          % 4th quadrant
  Y < 0.0,
  X >= - Y,         % angle >= 315 degrees
  !,
  Ratio is - Y / X,
  arctan2(Z1,Ratio),
  Z is - Z1.
arctan1(Z,X,Y) :-
  X > 0.0,          % 4th quadrant
  Y < 0.0,
  X < - Y,          % angle < 315 degrees
  !,
  Ratio is - X / Y,
  arctan2(Z1,Ratio),
  pi(Pi),
  Z is Z1 - (Pi / 2.0).


% arctan2(Y,X): Y is the inverse tangent of X, where 0.0 < X <= 1.0.
arctan2(Y,1.0) :-
  !,
  pi(Pi),
  Y is Pi / 4.0.
arctan2(Y,X) :-         % 0.0 < X < 1.0
  arctanSeriesLimit(L),
  arctanSeries(X,X,Y,1,1,L).


% arctanSeries(X,Term,Y,Sign,Exp,Limit): Y is the inverse tangent
%   of X (0.0 < X < 1.0) according to the series:
%
%   real__isArctanOf_real(X) = X - X^3/3 + X^5/5 - X^7/7 + X^9/9 - ... +/-
%                          (1/Exp)*X^Exp ]
%
%      where Exp is the largest odd <= Limit.


arctanSeries(_,_,0,_,Exp,Limit) :-
  Exp > Limit,
  !.
arctanSeries(X,Term,Y,Sign,Exp,Limit) :-
  Term1 is Term * X * X,
  Term2 is Sign * (Term / Exp),
  Exp2 is Exp + 2,
  OppSign is Sign * (-1),
  arctanSeries(X,Term1,Y2,OppSign,Exp2,Limit),
  Y is Term2 + Y2.



pi(3.141592654).
sqrtPrecision(1.0E-10).  % decimal-place precision of squareroot
expSeriesLimit(50).     % Limit on terms of exp series.
logSeriesLimit(99).     % Limit on terms of log series.
arctanSeriesLimit(99).  % Limit on terms of tan series.
sinSeriesLimit(50).     % Limit on terms of sin series.
cosSeriesLimit(99).     % Limit on terms of cos series.

